/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// screen.c -- master for refresh, status bar, console, chat, notify, etc

// mh 2010-07-16 - replaced with RMQ gl_screen.c
#include "quakedef.h"

/*

background clear
rendering
turtle/net/ram icons
sbar
centerprint / slow centerprint
notify lines
intermission / finale overlay
loading plaque
console
menu

required background clears
required update regions


syncronous draw mode or async
One off screen buffer, with updates either copied or xblited
Need to double buffer?


async draw will require the refresh area to be cleared, because it will be
xblited, but sync draw can just ignore it.

sync
draw

CenterPrint ()
SlowPrint ()
Screen_Update ();
Con_Printf ();

net
turn off messages option

the refresh is allways rendered, unless the console is full screen


console is:
	notify lines
	half
	full


*/


float		scr_con_current;
float		scr_conlines;		// lines of console to display

float		oldscreensize, oldfov, oldsbarscale, oldsbaralpha; // johnfitz -- added oldsbarscale and oldsbaralpha

// johnfitz -- new cvars
cvar_t		scr_menuscale = {"scr_menuscale", "1", true};
cvar_t		scr_sbarscale = {"scr_sbarscale", "1", true};
cvar_t		scr_sbaralpha = {"scr_sbaralpha", "1", true};
cvar_t		scr_conwidth = {"scr_conwidth", "0", true};
cvar_t		scr_conscale = {"scr_conscale", "1", true};
cvar_t		scr_scale = {"scr_scale", "1", true};
cvar_t		scr_crosshaircale = {"scr_crosshaircale", "1", true};
#ifdef _DEBUG
cvar_t		scr_showfps = {"scr_showfps", "1"}; // mh 2010-12-19 - reverted to off for release
#else
cvar_t		scr_showfps = {"scr_showfps", "0"}; // mh 2010-12-19 - reverted to off for release
#endif
cvar_t		scr_clock = {"scr_clock", "0"};
// johnfitz

cvar_t		scr_widescreenfov = {"scr_widescreenfov", "1", true};
cvar_t		scr_viewsize = {"viewsize", "100", true};
cvar_t		scr_fov = {"fov", "90"};	// 10 - 170

// mh 2010-07-16 - changed default to 3000 from 500, no longer saves to config
cvar_t		scr_conspeed = {"scr_conspeed", "3000"};

cvar_t		scr_centertime = {"scr_centertime", "2"};
cvar_t		scr_showram = {"showram", "1"};
cvar_t		scr_showturtle = {"showturtle", "0"};
cvar_t		scr_showpause = {"showpause", "1"};
cvar_t		scr_printspeed = {"scr_printspeed", "8"};
cvar_t		gl_triplebuffer = {"gl_triplebuffer", "1", true };

cvar_t		crosshaircolor = {"crosshaircolor", "109", true};

extern	cvar_t	crosshair;

qboolean	scr_initialized;		// ready to draw

qpic_t		*scr_ram;
qpic_t		*scr_net;
qpic_t		*scr_turtle;

int			clearconsole;
int			clearnotify;

qboolean	scr_disabled_for_loading;
qboolean	scr_drawloading;
float		scr_disabled_time;

qboolean	block_drawing;

qboolean	scr_incapture = false;
char		scr_capturename[MAX_OSPATH];
int			scr_captureframe;
int			scr_capturewidth;
int			scr_captureheight;
int			scr_captureperiod;

// 4:3 ratio default; downsampling prevents disk IO bottlenecks
cvar_t		cv_scr_capturewidth = {"scr_capturewidth", "320"};
cvar_t		cv_scr_captureheight = {"scr_captureheight", "240"};

cvar_t		scr_capturefps = {"scr_capturefps", "25"};

// buffer for capture
byte *scr_capturebuf = NULL;
byte *scr_scaledbuf = NULL;

void SCR_ScreenShot_f (void);

void Draw_End (void);

/*
===============================================================================

CENTER PRINTING

===============================================================================
*/

char		scr_centerstring[1024];
float		scr_centertime_start;	// for slow victory printing
float		scr_centertime_off;
int			scr_center_lines;
int			scr_erase_lines;
int			scr_erase_center;

/*
==============
SCR_CenterPrint

Called for important messages that should stay in the center of the screen
for a few moments
==============
*/
void SCR_CenterPrint (char *str) // update centerprint data
{
	strncpy (scr_centerstring, str, sizeof (scr_centerstring) - 1);
	scr_centertime_off = scr_centertime.value;
	scr_centertime_start = cl.time;

	// count the number of lines for centering
	scr_center_lines = 1;

	while (*str)
	{
		if (*str == '\n')
			scr_center_lines++;

		str++;
	}
}

void SCR_DrawCenterString (void) // actually do the drawing
{
	char	*start;
	int		l;
	int		j;
	int		x, y;
	int		remaining;

	GL_Set2DOffsets (OFS2D_MENU);

	// the finale prints the characters one at a time
	if (cl.intermission)
		remaining = scr_printspeed.value * (cl.time - scr_centertime_start);
	else remaining = 9999;

	scr_erase_center = 0;
	start = scr_centerstring;

	// hack for intermission text
	if (cl.intermission == 2 && key_dest == key_game)
		y = 32;
	else y = 200 * 0.10 - scr_center_lines * 4;

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
			if (start[l] == '\n' || !start[l])
				break;

		x = (320 - l * 8) >> 1;	// johnfitz -- 320x200 coordinate system

		for (j = 0; j < l; j++, x += 8)
		{
			Draw_Character (x, y, start[j]);	// johnfitz -- stretch overlays

			if (!remaining--)
				return;
		}

		y += 8;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;

		start++;		// skip the \n
	} while (1);
}

void SCR_CheckDrawCenterString (void)
{
	if (scr_center_lines > scr_erase_lines)
		scr_erase_lines = scr_center_lines;

	scr_centertime_off -= cl.frametime;

	if (scr_centertime_off <= 0 && !cl.intermission)
		return;

	if (key_dest != key_game)
		return;

	if (cl.paused) // johnfitz -- don't show centerprint during a pause
		return;

	SCR_DrawCenterString ();
}

// =============================================================================

/*
====================
CalcFovY
====================
*/
float SCR_CalcFovX (float fov_y, float width, float height)
{
	float   a;
	float   y;

	// bound, don't crash
	if (fov_y < 1) fov_y = 1;
	if (fov_y > 179) fov_y = 179;

	y = height / tan (fov_y / 360 * M_PI);
	a = atan (width / y);
	a = a * 360 / M_PI;

	return a;
}


float SCR_CalcFovY (float fov_x, float width, float height)
{
	float   a;
	float   x;

	// bound, don't crash
	if (fov_x < 1) fov_x = 1;
	if (fov_x > 179) fov_x = 179;

	x = width / tan (fov_x / 360 * M_PI);
	a = atan (height / x);
	a = a * 360 / M_PI;

	return a;
}


float CalcFovY (float fov_x, float width, float height)
{
	float   a, x;

	if (fov_x < 1 || fov_x > 179)
		Sys_Error ("Bad fov: %f", fov_x);

	x = width / tan (fov_x / 360 * M_PI);

	a = atan (height / x);
	a = a * 360 / M_PI;

	return a;
}


void SCR_CalcFOV (refdef_t *rd, float fov)
{
	rd->fov_x = fov;

	if (!scr_widescreenfov.value)
	{
		rd->fov_y = CalcFovY (rd->fov_x, vid.glwidth, vid.glheight - vid.sblines);
	}
	else
	{
		// calculate y fov as if the screen was 640 x 432; this ensures that the top and bottom
		// don't get clipped off if we have a widescreen display
		rd->fov_y = SCR_CalcFovY (rd->fov_x, 640, 432);

		// now recalculate fov_x so that it's correctly proportioned for fov_y
		rd->fov_x = SCR_CalcFovX (rd->fov_y, vid.glwidth, vid.glheight - vid.sblines);
	}
}


/*
=================
SCR_CalcRefdef

Must be called whenever vid changes
Internal use only

GL_BeginRendering must be called before this is called
=================
*/
static void SCR_CalcRefdef (void)
{
	float		scale; // johnfitz -- scale

	vid.recalc_refdef = 0;

	// bound viewsize
	if (scr_viewsize.value < 30) Cvar_Set ("viewsize", "30");
	if (scr_viewsize.value > 120) Cvar_Set ("viewsize", "120");

	// bound fov
	if (scr_fov.value < 10) Cvar_Set ("fov", "10");
	if (scr_fov.value > 170) Cvar_Set ("fov", "170");

	// johnfitz -- rewrote this section
	scale = CLAMP (1.0, scr_sbarscale.value, (float) vid.glwidth / 320.0);

	if (scr_viewsize.value >= 120 || cl.intermission || scr_sbaralpha.value < 1) // johnfitz -- scr_sbaralpha.value
		vid.sblines = 0;
	else if (scr_viewsize.value >= 110)
		vid.sblines = 24 * scale;
	else vid.sblines = 48 * scale;

	SCR_CalcFOV (&r_refdef, scr_fov.value);
}


/*
=================
SCR_SizeUp_f

Keybinding command
=================
*/
void SCR_SizeUp_f (void)
{
	Cvar_SetValue ("viewsize", scr_viewsize.value + 10);
	vid.recalc_refdef = 1;
}


/*
=================
SCR_SizeDown_f

Keybinding command
=================
*/
void SCR_SizeDown_f (void)
{
	Cvar_SetValue ("viewsize", scr_viewsize.value - 10);
	vid.recalc_refdef = 1;
}

/*
==================
SCR_Conwidth_f -- johnfitz -- called when scr_conwidth or scr_conscale changes
==================
*/
void SCR_Conwidth_f (void)
{
	vid.conwidth = (scr_conwidth.value > 0) ? (int) scr_conwidth.value : (scr_conscale.value > 0) ? (int) (vid.width / scr_conscale.value) : vid.width;
	vid.conwidth = CLAMP (320, vid.conwidth, vid.width);
	vid.conwidth &= 0xFFFFFFF8;
	vid.conheight = vid.conwidth * vid.height / vid.width;
}

// ============================================================================

/*
==================
SCR_LoadPics -- johnfitz
==================
*/
void SCR_LoadPics (void)
{
	scr_ram = Draw_PicFromWad ("ram");
	scr_net = Draw_PicFromWad ("net");
	scr_turtle = Draw_PicFromWad ("turtle");
}


void SCR_ForceRefdefRecalc (void)
{
	extern qboolean r_rebuildworld;

	// more sensible and sane way of doing this (rather than handling multiple "old" versions...)
	vid.recalc_refdef = 1;

	// force surface recache as more/less may be visible
	r_rebuildworld = true;
}


void CaptureBeginAudio (char *capturename);

void SCR_BeginCapture (char *capturedir)
{
	char name[MAX_OSPATH];

	sprintf (name, "%s/capture", com_gamedir);
	Sys_mkdir (name);

	sprintf (name, "%s/capture/%s", com_gamedir, capturedir);
	Sys_mkdir (name);

	Con_Printf ("Capturing to %s...\n", name);

	strcpy (scr_capturename, capturedir);
	scr_incapture = true;
	scr_captureframe = 0;

	// these should not be changed once a capture is running so store them out now
	scr_capturewidth = cv_scr_capturewidth.value;
	scr_captureheight = cv_scr_captureheight.value;

	if (scr_capturefps.value < 1)
		scr_captureperiod = 1000;
	else if (scr_capturefps.value > 500)
		scr_captureperiod = 1;
	else scr_captureperiod = 1000 / scr_capturefps.value;

	CaptureBeginAudio (scr_capturename);
}


void SCR_Capture_f (void)
{
	int argc = Cmd_Argc ();
	char *cmd = NULL;

	if (argc < 2)
		goto badargs;

	cmd = Cmd_Argv (1);

	if (!strcmp (cmd, "start"))
	{
		if (scr_incapture)
			Con_Printf ("Already capturing\n");
		else if (argc < 3)
			goto badargs;
		else SCR_BeginCapture (Cmd_Argv (2));
	}
	else if (!strcmp (cmd, "stop"))
	{
		if (scr_incapture)
		{
			scr_incapture = false;
			Con_Printf ("Capture stopped\n");
		}
		else Con_Printf ("Cannot stop - not currently recording\n");
	}
	else goto badargs;

	return;

badargs:;
	Con_Printf ("capture start <directory> : capture AV playback to the specified directory\n");
	Con_Printf ("capture stop : stop capturing\n");
}


/*
==================
SCR_Init
==================
*/
void SCR_Init (void)
{
	// capture
	Cvar_RegisterVariable (&cv_scr_capturewidth, NULL);
	Cvar_RegisterVariable (&cv_scr_captureheight, NULL);
	Cvar_RegisterVariable (&scr_capturefps, NULL);

	Cvar_RegisterVariable (&crosshaircolor, NULL);

	// johnfitz -- new cvars
	Cvar_RegisterVariable (&scr_menuscale, NULL);
	Cvar_RegisterVariable (&scr_sbarscale, SCR_ForceRefdefRecalc);
	Cvar_RegisterVariable (&scr_sbaralpha, SCR_ForceRefdefRecalc);
	Cvar_RegisterVariable (&scr_conwidth, &SCR_Conwidth_f);
	Cvar_RegisterVariable (&scr_conscale, &SCR_Conwidth_f);
	Cvar_RegisterVariable (&scr_scale, NULL);
	Cvar_RegisterVariable (&scr_crosshaircale, NULL);
	Cvar_RegisterVariable (&scr_showfps, NULL);
	Cvar_RegisterVariable (&scr_clock, NULL);
	// johnfitz

	// these all need a refdef recalc
	Cvar_RegisterVariable (&scr_widescreenfov, SCR_ForceRefdefRecalc);
	Cvar_RegisterVariable (&scr_fov, SCR_ForceRefdefRecalc);
	Cvar_RegisterVariable (&scr_viewsize, SCR_ForceRefdefRecalc);
	Cvar_RegisterVariable (&scr_conspeed, NULL);
	Cvar_RegisterVariable (&scr_showram, NULL);
	Cvar_RegisterVariable (&scr_showturtle, NULL);
	Cvar_RegisterVariable (&scr_showpause, NULL);
	Cvar_RegisterVariable (&scr_centertime, NULL);
	Cvar_RegisterVariable (&scr_printspeed, NULL);
	Cvar_RegisterVariable (&gl_triplebuffer, NULL);

	Cmd_AddCommand ("capture", SCR_Capture_f);
	Cmd_AddCommand ("screenshot", SCR_ScreenShot_f);
	Cmd_AddCommand ("sizeup", SCR_SizeUp_f);
	Cmd_AddCommand ("sizedown", SCR_SizeDown_f);

	SCR_LoadPics (); // johnfitz

	scr_initialized = true;
}

// ============================================================================

/*
==============
SCR_DrawFPS -- johnfitz
==============
*/
// this is incremented for each swapbuffers call giving us a proper framecount
int scr_framecount = 0;

void SCR_DrawFPS (void)
{
	static float	oldtime = 0, fps = 0;
	static int		oldframecount = 0;
	float			time;
	char			str[12];
	int				x, y, frames;

	time = realtime - oldtime;
	frames = scr_framecount - oldframecount;

	if (time < 0 || frames < 0)
	{
		oldtime = realtime;
		oldframecount = scr_framecount;
		return;
	}

	if (time > 0.25) // update value every 1/4 second
	{
		fps = frames / time;
		oldtime = realtime;
		oldframecount = scr_framecount;
	}

	if (scr_showfps.value) // draw it
	{
		sprintf (str, "%4.0f fps", fps);
		x = 320 - (strlen (str) << 3) - 4;
		y = 200 - 12;

		if (scr_clock.value) y -= 8; // make room for clock

		GL_Set2DOffsets (OFS2D_BOTTOMRIGHT);
		Draw_String (x, y, str);
	}
}


/*
==============
SCR_DrawClock -- johnfitz
==============
*/
void SCR_DrawClock (void)
{
	char	str[9];

	if (scr_clock.value == 1)
	{
		int minutes, seconds;

		minutes = cl.time / 60;
		seconds = ((int) cl.time) % 60;

		sprintf (str, "%i:%i%i", minutes, seconds / 10, seconds % 10);
	}
	else
		return;

	// draw it
	GL_Set2DOffsets (OFS2D_BOTTOMRIGHT);
	Draw_String (320 - (strlen (str) << 3), 200 - 8, str);
}


/*
==============
SCR_DrawDevStats
==============
*/
void SCR_DrawDevStats (void)
{
	char	str[40];
	int		y = 25 - 9; // 9=number of lines to print
	int		x = 0; // margin

	if (!devstats.value)
		return;

	GL_Set2DOffsets (OFS2D_BOTTOMLEFT);

	Draw_Fill (x, y * 8, 19 * 8, 9 * 8, 0, 0.5); // dark rectangle

	sprintf (str, "devstats |Curr Peak");
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "---------+---------");
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Edicts   |%4i %4i", dev_stats.edicts, dev_peakstats.edicts);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Packet   |%4i %4i", dev_stats.packetsize, dev_peakstats.packetsize);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Visedicts|%4i %4i", dev_stats.visedicts, dev_peakstats.visedicts);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Efrags   |%4i %4i", dev_stats.efrags, dev_peakstats.efrags);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Dlights  |%4i %4i", dev_stats.dlights, dev_peakstats.dlights);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Beams    |%4i %4i", dev_stats.beams, dev_peakstats.beams);
	Draw_String (x, (y++) * 8 - x, str);

	sprintf (str, "Tempents |%4i %4i", dev_stats.tempents, dev_peakstats.tempents);
	Draw_String (x, (y++) * 8 - x, str);
}

/*
==============
SCR_DrawRam
==============
*/
void SCR_DrawRam (void)
{
	if (!scr_showram.value)
		return;

	if (!r_cache_thrash)
		return;

	GL_Set2DOffsets (OFS2D_DEFAULT); // johnfitz
	Draw_Pic (32, 0, scr_ram, DP_DEFAULT);
}

/*
==============
SCR_DrawTurtle
==============
*/
void SCR_DrawTurtle (void)
{
	static int	count;

	if (!scr_showturtle.value)
		return;

	if (cl.frametime < 0.1)
	{
		count = 0;
		return;
	}

	count++;

	if (count < 3)
		return;

	GL_Set2DOffsets (OFS2D_DEFAULT); // johnfitz
	Draw_Pic (0, 0, scr_turtle, DP_DEFAULT);
}

/*
==============
SCR_DrawNet
==============
*/
void SCR_DrawNet (void)
{
	if (realtime - cl.last_received_message < 0.3)
		return;

	if (cls.demoplayback)
		return;

	GL_Set2DOffsets (OFS2D_DEFAULT); // johnfitz
	Draw_Pic (64, 0, scr_net, DP_DEFAULT);
}


/*
==============
DrawPause
==============
*/
void SCR_DrawPause (void)
{
	qpic_t	*pic;

	if (!cl.paused)
		return;

	if (!scr_showpause.value)		// turn off for screenshots
		return;

	GL_Set2DOffsets (OFS2D_MENU); // johnfitz

	pic = Draw_CachePic ("gfx/pause.lmp");
	Draw_Pic ((320 - pic->width) / 2, (240 - 48 - pic->height) / 2, pic, DP_DEFAULT); // johnfitz -- stretched menus
}


/*
==============
SCR_DrawLoading
==============
*/
void SCR_DrawLoading (void)
{
	qpic_t	*pic;

	if (!scr_drawloading)
		return;

	GL_Set2DOffsets (OFS2D_MENU); // johnfitz

	pic = Draw_CachePic ("gfx/loading.lmp");
	Draw_Pic ((320 - pic->width) / 2, (240 - 48 - pic->height) / 2, pic, DP_DEFAULT); // johnfitz -- stretched menus
}


/*
==============
SCR_DrawCrosshair -- johnfitz
==============
*/
void SCR_DrawCrosshair (void)
{
	int crossx;
	int crossy;
	float s = CLAMP (1.0, scr_crosshaircale.value, 10.0);
	extern gltexture_t *new_crosshair;

	if (!crosshair.value || cl.qc.hidecrosshair)
		return;

	if (crosshair.value > 1)
	{
		// was 4, 8 for new crosshair
		crossx = ((vid.glwidth / 2) - (8.0f * s)) / s;
		crossy = (((vid.glheight - vid.sblines) / 2) - (8.0f * s)) / s;
	}
	else
	{
		// was 4 for old crosshair
		crossx = ((vid.glwidth / 2) - (4.0f * s)) / s;
		crossy = (((vid.glheight - vid.sblines) / 2) - (4.0f * s)) / s;
	}

	GL_Set2DOffsets (OFS2D_DEFAULT);

	if (s != 1)
	{
		glPushMatrix ();
		glScalef (s, s, 1);
	}

	if (crosshair.value > 1)
		Draw_Texture (crossx, crossy, 16, 16, new_crosshair, (byte *) &d_8to24table_rgba[(int) crosshaircolor.value & 255]);
	else Draw_Character (crossx, crossy, '+');

	if (s != 1) glPopMatrix ();
}



// =============================================================================


/*
==================
SCR_SetUpToDrawConsole
==================
*/
void SCR_SetUpToDrawConsole (void)
{
	// C doesn't allow us to init statics with a variable. :(
	static double oldcontime = -1;
	double frametime;

	// so we do it this way instead
	if (oldcontime < 0) oldcontime = realtime;

	// get the host_timescale/host_framerate independent time
	frametime = realtime - oldcontime;
	oldcontime = frametime;

	Con_CheckResize ();

	if (scr_drawloading)
		return;		// never a console with loading plaque

	// decide on the height of the console
	con_forcedup = !cl.worldmodel || cls.signon != SIGNONS;

	if (con_forcedup)
	{
		scr_conlines = vid.glheight; // full screen // johnfitz -- vid.glheight instead of vid.height
		scr_con_current = scr_conlines;
	}
	else if (key_dest == key_console)
		scr_conlines = vid.glheight / 2; // half screen // johnfitz -- vid.glheight instead of vid.height
	else
		scr_conlines = 0; // none visible

	if (scr_conlines < scr_con_current)
	{
		scr_con_current -= scr_conspeed.value * frametime;

		if (scr_conlines > scr_con_current)
			scr_con_current = scr_conlines;

	}
	else if (scr_conlines > scr_con_current)
	{
		scr_con_current += scr_conspeed.value * frametime;

		if (scr_conlines < scr_con_current)
			scr_con_current = scr_conlines;
	}
}


/*
==================
SCR_DrawConsole
==================
*/
void SCR_DrawConsole (void)
{
	if (scr_con_current)
	{
		Con_DrawConsole (scr_con_current, true);
		clearconsole = 0;
	}
	else
	{
		if (key_dest == key_game || key_dest == key_message)
			Con_DrawNotify ();	// only draw notify in game
	}
}


/*
==============================================================================

						SCREEN SHOTS

==============================================================================
*/

/*
==================
SCR_ScreenShot_f -- johnfitz -- rewritten to use Image_WriteTGA
==================
*/
void SCR_ScreenShot_f (void)
{
	byte	*buffer;
	char	tganame[MAX_OSPATH];
	char	checkname[MAX_OSPATH];
	int	i;

	sprintf (checkname, "%s/screenshot", com_gamedir);
	Sys_mkdir (checkname);

	// find a file name to save it to
	for (i = 0; i < 10000; i++)
	{
		sprintf (tganame, "screenshot/rmq%04i.tga", i);	// "fitz%04i.tga"
		sprintf (checkname, "%s/%s", com_gamedir, tganame);

		if (Sys_FileTime (checkname) == -1)
			break;	// file doesn't exist
	}

	if (i == 10000)
	{
		Con_Printf ("SCR_ScreenShot_f: Couldn't find an unused filename\n");
		return;
	}

	// get data
	buffer = (byte *) malloc (vid.glwidth * vid.glheight * 3);
	glReadPixels (0, 0, vid.glwidth, vid.glheight, GL_BGR, GL_UNSIGNED_BYTE, buffer);

	// now write the file
	if (Image_WriteTGA (tganame, buffer, vid.glwidth, vid.glheight, 24, false))
		Con_Printf ("Wrote %s\n", tganame);
	else Con_Printf ("SCR_ScreenShot_f: Couldn't create a TGA file\n");

	free (buffer);
}


// =============================================================================


/*
===============
SCR_BeginLoadingPlaque

================
*/
void SCR_BeginLoadingPlaque (void)
{
	S_StopAllSounds (true);

	if (cls.state != ca_connected)
		return;

	if (cls.signon != SIGNONS)
		return;

	// redraw with no console and the loading plaque
	Con_ClearNotify ();
	scr_centertime_off = 0;
	scr_con_current = 0;

	scr_drawloading = true;
	SCR_UpdateScreen ();
	scr_drawloading = false;

	scr_disabled_for_loading = true;
	scr_disabled_time = realtime;
}

/*
===============
SCR_EndLoadingPlaque

================
*/
void SCR_EndLoadingPlaque (void)
{
	scr_disabled_for_loading = false;
	Con_ClearNotify ();
}

// =============================================================================

char	*scr_notifystring;
qboolean	scr_drawdialog;

void SCR_DrawNotifyString (void)
{
	char	*start;
	int		l;
	int		j;
	int		x, y;

	GL_Set2DOffsets (OFS2D_MENU); // johnfitz

	start = scr_notifystring;

	y = 200 * 0.35; // johnfitz -- stretched overlays

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
			if (start[l] == '\n' || !start[l])
				break;

		x = (320 - l * 8) / 2; // johnfitz -- stretched overlays

		for (j = 0; j < l; j++, x += 8)
			Draw_Character (x, y, start[j]);

		y += 8;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;

		start++;		// skip the \n
	}
	while (1);
}

/*
==================
SCR_ModalMessage

Displays a text string in the center of the screen and waits for a Y or N
keypress.
==================
*/
int SCR_ModalMessage (char *text, float timeout) // johnfitz -- timeout
{
	int time1, time2; // johnfitz -- timeout

	if (cls.state == ca_dedicated)
		return true;

	scr_notifystring = text;

	// draw a fresh screen
	scr_drawdialog = true;
	SCR_UpdateScreen ();
	scr_drawdialog = false;

	S_ClearBuffer ();		// so dma doesn't loop current sound

	time1 = SDL_GetTicks () + (int) (timeout * 1000); // johnfitz -- timeout
	time2 = 0; // johnfitz -- timeout

	do
	{
		key_count = -1;		// wait for a key down and up
		Sys_SendKeyEvents ();

		if (timeout > 0) time2 = SDL_GetTicks (); // johnfitz -- zero timeout means wait forever.
	} while (key_lastpress != 'y' &&
			key_lastpress != 'n' &&
			key_lastpress != K_ESCAPE &&
			time2 <= time1);

	// 	SCR_UpdateScreen (); // johnfitz -- commented out

	// johnfitz -- timeout
	if (time2 > time1)
		return false;

	// johnfitz

	return key_lastpress == 'y';
}


// =============================================================================

// johnfitz -- deleted SCR_BringDownConsole


void GL_Resample32BitTexture (unsigned *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight)
{
	int		i, j;
	unsigned	*inrow, *inrow2;
	unsigned	frac, fracstep;
	unsigned	*p1, *p2;
	byte		*pix1, *pix2, *pix3, *pix4;

	p1 = (unsigned *) Hunk_Alloc (outwidth * 4);
	p2 = (unsigned *) Hunk_Alloc (outwidth * 4);

	fracstep = inwidth * 0x10000 / outwidth;

	frac = fracstep >> 2;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	frac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	for (i = 0; i < outheight; i++, out += outwidth)
	{
		inrow = in + inwidth * (int) ((i + 0.25) * inheight / outheight);
		inrow2 = in + inwidth * (int) ((i + 0.75) * inheight / outheight);
		frac = fracstep >> 1;

		for (j = 0; j < outwidth; j++)
		{
			pix1 = (byte *) inrow + p1[j];
			pix2 = (byte *) inrow + p2[j];
			pix3 = (byte *) inrow2 + p1[j];
			pix4 = (byte *) inrow2 + p2[j];

			((byte *) (out + j))[0] = (pix1[0] + pix2[0] + pix3[0] + pix4[0]) >> 2;
			((byte *) (out + j))[1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1]) >> 2;
			((byte *) (out + j))[2] = (pix1[2] + pix2[2] + pix3[2] + pix4[2]) >> 2;
			((byte *) (out + j))[3] = (pix1[3] + pix2[3] + pix3[3] + pix4[3]) >> 2;
		}
	}
}


/*
==================
SCR_UpdateScreen

This is called every frame, and can also be called explicitly to flush
text to the screen.

WARNING: be very careful calling this from elsewhere, because the refresh
needs almost the entire 256k of stack space!
==================
*/
void TM_MipMapWH (unsigned *data, int width, int height);
unsigned *TM_MipMapW (unsigned *data, int width, int height);
unsigned *TM_MipMapH (unsigned *data, int width, int height);


void SCR_UpdateScreen (void)
{
	if (block_drawing)
		return;

	vid.numpages = 2 + (gl_triplebuffer.value ? 1 : 0); // johnfitz -- in case gl_triplebuffer is not 0 or 1

	if (scr_disabled_for_loading)
	{
		if (realtime - scr_disabled_time > 60)
		{
			scr_disabled_for_loading = false;
			Con_Printf ("load failed.\n");
		}
		else return;
	}

	if (!scr_initialized || !con_initialized)
		return;				// not initialized yet

	// unbind everything
	GL_UncacheArrays ();
	GL_UnbindTextures ();
	GL_SetVertexProgram (0);
	GL_SetFragmentProgram (0);

	GL_BeginRendering (&vid);

	// determine size of refresh window
	if (vid.recalc_refdef)
		SCR_CalcRefdef ();

	// do 3D refresh drawing, and then update the screen
	SCR_SetUpToDrawConsole ();

	if (!CL_CSQC_UpdateView ())
	{
		V_RenderView ();
		GL_Set2D ();
	}

	if (scr_drawdialog) // new game confirm
	{
		Sbar_Draw ();
		Draw_FadeScreen ();
		SCR_DrawNotifyString ();
	}
	else if (scr_drawloading) // loading
	{
		SCR_DrawLoading ();
		Sbar_Draw ();
	}
	else if (cl.intermission == 1 && key_dest == key_game) // end of level
	{
		Sbar_IntermissionOverlay ();
	}
	else if (cl.intermission == 2 && key_dest == key_game) // end of episode
	{
		Sbar_FinaleOverlay ();
		SCR_CheckDrawCenterString ();
	}
	else
	{
		SCR_DrawCrosshair (); // johnfitz
		SCR_DrawRam ();
		SCR_DrawNet ();
		SCR_DrawTurtle ();
		SCR_DrawPause ();
		SCR_CheckDrawCenterString ();
		Sbar_Draw ();
		SCR_DrawDevStats (); // johnfitz
		SCR_DrawFPS (); // johnfitz
		SCR_DrawClock (); // johnfitz
		SCR_DrawConsole ();
		M_Draw ();
	}

	Draw_End ();
	V_UpdateBlend (); // johnfitz -- V_UpdatePalette cleaned up and renamed

	GL_EndRendering ();

	if (scr_incapture)
	{
		char tganame[MAX_OSPATH];
		byte *scaled = NULL;

		if (!scr_capturebuf) scr_capturebuf = (byte *) malloc (vid.glwidth * vid.glheight * 4);

		// read as quickly as possible; it's going to stall the pipeline *anyway* but let's not make things any worse than they already are
		glReadPixels (0, 0, vid.glwidth, vid.glheight, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, scr_capturebuf);

		// bound sensible
		if (scr_capturewidth > vid.glwidth) scr_capturewidth = vid.glwidth;
		if (scr_captureheight > vid.glheight) scr_captureheight = vid.glheight;
		if (scr_capturewidth < 100) scr_capturewidth = 100;
		if (scr_captureheight < 100) scr_captureheight = 100;

		if (vid.glwidth == scr_capturewidth && vid.glheight == scr_captureheight)
			scaled = scr_capturebuf;
		else
		{
			int srcwidth = vid.glwidth;
			int srcheight = vid.glheight;

			// mipmap down in-place if > 2* (faster than resampling and preserves quality better)
			// should this be > or >= ???
			while (srcwidth >= scr_capturewidth * 2 && srcheight >= scr_captureheight * 2)
			{
				TM_MipMapWH ((unsigned *) scr_capturebuf, srcwidth, srcheight);
				srcwidth >>= 1;
				srcheight >>= 1;
			}

			while (srcwidth >= scr_capturewidth * 2)
			{
				TM_MipMapW ((unsigned *) scr_capturebuf, srcwidth, srcheight);
				srcwidth >>= 1;
			}

			while (srcheight >= scr_captureheight * 2)
			{
				TM_MipMapH ((unsigned *) scr_capturebuf, srcwidth, srcheight);
				srcheight >>= 1;
			}

			if (srcwidth == scr_capturewidth && srcheight == scr_captureheight)
			{
				// no need to resample
				scaled = scr_capturebuf;
			}
			else
			{
				if (!scr_scaledbuf) scr_scaledbuf = (byte *) malloc (scr_capturewidth * scr_captureheight * 4);

				GL_Resample32BitTexture ((unsigned *) scr_capturebuf, srcwidth, srcheight, (unsigned *) scr_scaledbuf, scr_capturewidth, scr_captureheight);
				scaled = scr_scaledbuf;
			}
		}

		sprintf (tganame, "capture/%s/cap%07i.tga", scr_capturename, scr_captureframe);
		Image_WriteTGA (tganame, scaled, scr_capturewidth, scr_captureheight, 32, false);

		scr_captureframe++;
	}
	else
	{
		if (scr_scaledbuf)
		{
			free (scr_scaledbuf);
			scr_scaledbuf = NULL;
		}

		if (scr_capturebuf)
		{
			free (scr_capturebuf);
			scr_capturebuf = NULL;
		}
	}
}

