/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// gl_warp.c -- warping animation support

#include "quakedef.h"

cvar_t r_lavafog = {"r_lavafog", "0", true};
cvar_t r_slimefog = {"r_slimefog", "0", true};
cvar_t r_lavaalpha = {"r_lavaalpha", "1", true};
cvar_t r_telealpha = {"r_telealpha", "1", true};
cvar_t r_slimealpha = {"r_slimealpha", "1", true};
cvar_t r_lockalpha = {"r_lockalpha", "0", true};
cvar_t r_warpfactor = {"r_warpfactor", "2", true};
cvar_t r_warpscale = {"r_warpscale", "8", true};
cvar_t r_warpspeed = {"r_warpspeed", "4", true};

void R_WarpRegisterCvars (void)
{
	Cvar_RegisterVariable (&r_warpspeed, NULL);
	Cvar_RegisterVariable (&r_warpscale, NULL);
	Cvar_RegisterVariable (&r_warpfactor, NULL);
	Cvar_RegisterVariable (&r_slimefog, NULL);
	Cvar_RegisterVariable (&r_lavafog, NULL);
	Cvar_RegisterVariable (&r_lavaalpha, NULL);
	Cvar_RegisterVariable (&r_telealpha, NULL);
	Cvar_RegisterVariable (&r_slimealpha, NULL);
	Cvar_RegisterVariable (&r_lockalpha, NULL);
}

/*
OK, just so you know where these numbers come from, and so I don't get to waste another 2 hours calculating backwards...

From FitzQuake:

#define WARPCALC(s,t) ((s + turbsin[(int)((t*2)+(cl.time*(128.0/M_PI))) & 255]) * (1.0/64)) //johnfitz -- correct warp

Used when gl_subdivide_size is below a certain threahold.  Now, we warp per-pixel so we're always below, so we use this

Broken out, turnsin lookup replaced with sin and we get:

#define RDT (cl.time * (128 / M_PI))
#define BYTEANGLE (360.0 / 256.0)
#define ANGLERAD (M_PI / 180.0)
#define TIMEMULT (BYTEANGLE * ANGLERAD) // 0.02454369260617025967548940143187
#define WARPCALC(s, t) ((s + sin (((t * 2) + RDT) * TIMEMULT) * 8.0f) * (1.0 / 64))

vertex program.env[0].x stores RDT; this is optionally multiplied by a player-adjustable speed factor
vertex program.env[0].y stores 2, which is multiplied by t above; this is also player-adjustable
TIMEMULT is hardcoded in the vertex program to blah decimal places
fragment program.env[0].z stores 8.0f, also player-adjustable, brings -1...1 scale up to turbsin -8...8
1.0 / 64 is hard-coded in the fragment program to bring the final result back to the regular scale
*/

#define liquidvp \
	"!!ARBvp1.0\n" \
	TRANSFORMVERTEX ("result.position", "vertex.attrib[0]") \
	"SUB result.texcoord[0], vertex.attrib[1], program.env[0].z;\n" \
	"MAD result.texcoord[1], vertex.attrib[2], program.env[0].y, program.env[0].x;\n" \
	"END\n"

#define liquidfp \
	"!!ARBfp1.0\n" \
	"TEMP tc0, tc1;\n" \
	"TEX tc0, fragment.texcoord[1], texture[1], 2D;\n" \
	"MAD tc0, tc0, program.env[0].w, fragment.texcoord[0];\n" \
	"TEX tc1, tc0, texture[0], 2D;\n" \
	"MUL result.color, tc1, program.env[1];\n" \
	"END\n"

GLuint arb_liquid_vp = 0;
GLuint arb_liquid_fp = 0;
GLuint arb_liquid_fp_fog = 0;


// this texture is for to provide a sine lookup to optimize the warp; it's not yet implemented in the shader
gltexture_t *r_warptexture = NULL;

byte Warp_SineWarp (int num, int detail)
{
	float ang = (num % detail);

	ang /= (float) detail;
	ang *= 360.0f;

	ang = sin (ang * (M_PI / 180.0f));
	ang += 1.0f;
	ang /= 2.0f;

	return BYTE_CLAMPF (ang);
}


byte Warp_EdgeWarp (int num, int detail)
{
	float ang = (num % detail);

	// 0..0 range
	ang /= (float) (detail - 1);
	ang *= 180.0f;

	// extreme clamp to force the bounding regions to the edges
	ang = sin (ang * (M_PI / 180.0f));
	ang *= 4.0f;

	return BYTE_CLAMPF (ang);
}


void Warp_CreateWarpTexture (void)
{
#define WARP_DETAIL	32
	int x, y;
	static byte data[WARP_DETAIL * WARP_DETAIL * 4];
	byte *bgra = data;

	for (y = 0; y < WARP_DETAIL; y++)
	{
		for (x = 0; x < WARP_DETAIL; x++, bgra += 4)
		{
			bgra[2] = Warp_SineWarp (x, WARP_DETAIL);
			bgra[1] = Warp_SineWarp (y, WARP_DETAIL);
			bgra[0] = Warp_EdgeWarp (x, WARP_DETAIL);
			bgra[3] = Warp_EdgeWarp (y, WARP_DETAIL);
		}
	}

	r_warptexture = TM_LoadImage (NULL, "warpcontrol", WARP_DETAIL, WARP_DETAIL, SRC_RGBA, (byte *) data, "", (src_offset_t) data, TEXPREF_NOPICMIP);
}


void Warp_CreateShaders (void)
{
	Warp_CreateWarpTexture ();

	arb_liquid_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, liquidvp, false);
	arb_liquid_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, liquidfp, false);
	arb_liquid_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, liquidfp, true);
}


void WaterAlpha_ParseWorldspawn (void)
{
	// did we agree that this wasn't being done???
}


cvar_t r_waterwarp = {"r_waterwarp", "1", true};

// cached states
float r_lastliquidalpha = -1;
float r_warpcurrentfog = -666;
gltexture_t *r_lastwarptexture = NULL;


void R_LiquidSetup (void)
{
	// common
	float params[4] = {0, 0, 0, 0};

	r_lastliquidalpha = -1;
	r_warpcurrentfog = -666;
	r_lastwarptexture = NULL;

	// these are env params for liquid because we need to toggle the fragment shader depending on fog state
	params[0] = cl.time * r_warpspeed.value * (0.125f / M_PI);
	params[1] = r_warpfactor.value * 0.00390625f;
	params[2] = r_warpscale.value * 0.015625f;
	params[3] = r_warpscale.value * 0.03125f;

	qglProgramEnvParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, params);
	qglProgramEnvParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 0, params);

	GL_BindTexture (GL_TEXTURE1, r_warptexture);
}


void R_BeginSurfaces (void);
void R_EndSurfaces (void);

void R_LiquidBegin (void)
{
	R_BeginSurfaces ();
	R_LiquidSetup ();
}


void R_LiquidTakeDown (void)
{
	R_FlushSurfaces ();

	// force to invalidate again
	r_lastliquidalpha = -1;
	r_warpcurrentfog = -666;
	r_lastwarptexture = NULL;

	glFogf (GL_FOG_DENSITY, Fog_GetDensity () / 64.0f);
}


void R_LiquidEnd (void)
{
	R_LiquidTakeDown ();
	R_EndSurfaces ();
}


void R_AddLiquidToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha);

void R_RenderLiquidSurface (msurface_t *surf, entity_t *ent, float surfalpha)
{
	float surffog;

	R_PrecheckSurface (surf, ent);

	surf->vertexprogram = arb_liquid_vp;
	surf->fragmentprogram = arb_liquid_fp;

	if (surf->flags & SURF_DRAWLAVA)
		surffog = Fog_GetDensity () * r_lavafog.value;
	else if (surf->flags & SURF_DRAWSLIME)
		surffog = Fog_GetDensity () * r_slimefog.value;
	else if (!(surf->flags & SURF_DRAWLAVA))
		surffog = Fog_GetDensity ();
	else if (!(surf->flags & SURF_DRAWSLIME))
		surffog = Fog_GetDensity ();
	else surffog = 0;

	if (surffog != r_warpcurrentfog)
	{
		R_FlushSurfaces ();

		if (surffog > 0)
		{
			glFogf (GL_FOG_DENSITY, surffog / 64.0f);
			surf->fragmentprogram = arb_liquid_fp_fog;
		}
		else surf->fragmentprogram = arb_liquid_fp;

		r_warpcurrentfog = surffog;
	}

	if (surfalpha != r_lastliquidalpha)
	{
		float alpha = surfalpha;

		if (alpha < 0) alpha = 0;
		if (alpha > 1) alpha = 1;

		R_FlushSurfaces ();
		qglProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 1, 1, 1, 1, alpha);
		r_lastliquidalpha = surfalpha;
	}

	if (surf->texinfo->texture->gltexture != r_lastwarptexture)
	{
		R_FlushSurfaces ();
		GL_BindTexture (GL_TEXTURE0, surf->texinfo->texture->gltexture);
		r_lastwarptexture = surf->texinfo->texture->gltexture;
	}

	R_BatchSurface (surf);
}


void R_DrawWaterSurfaces (msurface_t *chain, entity_t *ent, model_t *mod)
{
	int i;
	msurface_t *surf;
	qboolean stateset = false;

	// clear down the previous chains
	for (i = 0; i < mod->numtextures; i++)
	{
		texture_t *tex;

		if (!(tex = mod->textures[i])) continue;

		mod->textures[i]->texturechain = NULL;
	}

	// reverse the texture chain to front-to-back and sort by texture
	for (surf = chain; surf; surf = surf->lightmapchain)
	{
		surf->texturechain = surf->texinfo->texture->texturechain;
		surf->texinfo->texture->texturechain = surf;
	}

	for (i = 0; i < mod->numtextures; i++)
	{
		texture_t *tex;

		if (!(tex = mod->textures[i])) continue;
		if (!(surf = tex->texturechain)) continue;

		for (; surf; surf = surf->texturechain)
		{
			// override water alpha for certain surface types
			if (!r_lockalpha.value)
			{
				if (surf->flags & SURF_DRAWLAVA) surf->wateralpha = r_lavaalpha.value;
				if (surf->flags & SURF_DRAWTELE) surf->wateralpha = r_telealpha.value;
				if (surf->flags & SURF_DRAWSLIME) surf->wateralpha = r_slimealpha.value;
			}

			// check for alpha surfs and defer them to the alpha list
			if (surf->wateralpha < 1)
			{
				R_AddLiquidToAlphaList (surf, ent, surf->wateralpha);
				continue;
			}

			// done this way because we might be skipping some surfs if alpha
			if (!stateset)
			{
				R_LiquidSetup ();
				stateset = true;
			}

			// no alpha
			R_RenderLiquidSurface (surf, ent, 1.0f);
		}

		tex->texturechain = NULL;
	}

	if (stateset) R_LiquidTakeDown ();
}


