/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// host.c -- coordinates spawning and killing of local servers

#include "quakedef.h"

/*

A server can allways be started, even if the system started out as a client
to a remote system.

A client can NOT be started if the system started as a dedicated server.

Memory is cleared / released when a server or client begins, not when they end.

*/

quakeparms_t host_parms;

qboolean	host_initialized;		// true if into command execution

double		host_frametime;
double		host_lastframetime = 0;
double		realtime;				// without any filtering or bounding
double		oldrealtime;
double		nextframetime;
int			host_framecount;

int			host_hunklevel;

int			minimum_memory;

client_t	*host_client;			// current client

jmp_buf 	host_abortserver;

byte		*host_colormap;

cvar_t	host_framerate = {"host_framerate", "0"};	// set for slow motion
cvar_t	host_speeds = {"host_speeds", "0"};			// set for running times
cvar_t	host_maxfps = {"host_maxfps", "72", true}; // johnfitz
cvar_t	host_timescale = {"host_timescale", "0"}; // johnfitz

cvar_t	sys_ticrate = {"sys_ticrate", "0.05"}; // dedicated server
cvar_t	serverprofile = {"serverprofile", "0"};

cvar_t	fraglimit = {"fraglimit", "0", false, true};
cvar_t	timelimit = {"timelimit", "0", false, true};
cvar_t	teamplay = {"teamplay", "0", false, true};
cvar_t	samelevel = {"samelevel", "0"};
cvar_t	noexit = {"noexit", "0", false, true};
cvar_t	skill = {"skill", "1"};						// 0 - 3
cvar_t	deathmatch = {"deathmatch", "0"};			// 0, 1, or 2
cvar_t	coop = {"coop", "0"};			// 0 or 1

cvar_t	pausable = {"pausable", "1"};

cvar_t	developer = {"developer", "0"};

cvar_t	temp1 = {"temp1", "0"};
cvar_t	temp2 = {"temp2", "0"};
cvar_t	temp3 = {"temp3", "0"};
cvar_t	temp4 = {"temp4", "0"};

// mh - for testing various stuff that needs a cvar until we decide on the correct approach and either assign a cvar or hard-code the effect
cvar_t	test1 = {"test1", "0"};
cvar_t	test2 = {"test2", "0"};
cvar_t	test3 = {"test3", "0"};
cvar_t	test4 = {"test4", "0"};

cvar_t devstats = {"devstats", "0"}; // johnfitz -- track developer statistics that vary every frame

devstats_t dev_stats, dev_peakstats;
overflowtimes_t dev_overflows; // this stores the last time overflow messages were displayed, not the last time overflows occured


/*
================
Host_EndGame
================
*/
void Host_EndGame (const char *message, ...)
{
	va_list		argptr;
	char		string[1024];

	va_start (argptr, message);
	vsprintf (string, message, argptr);
	va_end (argptr);
	Con_DPrintf ("Host_EndGame: %s\n", string);

	if (sv.active)
		Host_ShutdownServer (false);

	if (cls.state == ca_dedicated)
		Sys_Error ("Host_EndGame: %s\n", string);	// dedicated servers exit

	if (cls.demonum != -1)
		CL_NextDemo ();
	else
		CL_Disconnect ();

	longjmp (host_abortserver, 1);
}

/*
================
Host_Error

This shuts down both the client and server
================
*/
void Host_Error (const char *error, ...)
{
	va_list		argptr;
	char		string[1024];
	static	qboolean inerror = false;

	if (inerror)
		Sys_Error ("Host_Error: recursively entered");

	inerror = true;

	SCR_EndLoadingPlaque ();		// reenable screen updates

	va_start (argptr, error);
	vsprintf (string, error, argptr);
	va_end (argptr);
	Con_Printf ("Host_Error: %s\n", string);

	if (sv.active)
		Host_ShutdownServer (false);

	if (cls.state == ca_dedicated)
		Sys_Error ("Host_Error: %s\n", string);	// dedicated servers exit

	CL_Disconnect ();
	cls.demonum = -1;
	cl.intermission = 0; // johnfitz -- for errors during intermissions (changelevel with no map found, etc.)

	inerror = false;

	longjmp (host_abortserver, 1);
}

/*
================
Host_FindMaxClients
================
*/
void	Host_FindMaxClients (void)
{
	int		i;

	svs.maxclients = 1;

	i = COM_CheckParm ("-dedicated");

	if (i)
	{
		cls.state = ca_dedicated;

		if (i != (com_argc - 1))
		{
			svs.maxclients = Q_atoi (com_argv[i+1]);
		}
		else
			svs.maxclients = 8;
	}
	else
		cls.state = ca_disconnected;

	i = COM_CheckParm ("-listen");

	if (i)
	{
		if (cls.state == ca_dedicated)
			Sys_Error ("Only one of -dedicated or -listen can be specified");

		if (i != (com_argc - 1))
			svs.maxclients = Q_atoi (com_argv[i+1]);
		else
			svs.maxclients = 8;
	}

	if (svs.maxclients < 1)
		svs.maxclients = 8;
	else if (svs.maxclients > MAX_SCOREBOARD)
		svs.maxclients = MAX_SCOREBOARD;

	svs.maxclientslimit = svs.maxclients;

	if (svs.maxclientslimit < 4)
		svs.maxclientslimit = 4;

	svs.clients = (struct client_s *) Hunk_AllocName (svs.maxclientslimit * sizeof (client_t), "clients");

	if (svs.maxclients > 1)
		Cvar_SetValue ("deathmatch", 1.0);
	else Cvar_SetValue ("deathmatch", 0.0);
}


/*
=======================
Host_InitLocal
======================
*/
void Host_InitLocal (void)
{
	Host_InitCommands ();

	Cvar_RegisterVariable (&host_framerate, NULL);
	Cvar_RegisterVariable (&host_speeds, NULL);
	Cvar_RegisterVariable (&host_maxfps, NULL); // johnfitz
	Cvar_RegisterVariable (&host_timescale, NULL); // johnfitz

	Cvar_RegisterVariable (&devstats, NULL); // johnfitz

	Cvar_RegisterVariable (&sys_ticrate, NULL);
	Cvar_RegisterVariable (&serverprofile, NULL);

	Cvar_RegisterVariable (&fraglimit, NULL);
	Cvar_RegisterVariable (&timelimit, NULL);
	Cvar_RegisterVariable (&teamplay, NULL);
	Cvar_RegisterVariable (&samelevel, NULL);
	Cvar_RegisterVariable (&noexit, NULL);
	Cvar_RegisterVariable (&skill, NULL);
	Cvar_RegisterVariable (&developer, NULL);
	Cvar_RegisterVariable (&deathmatch, NULL);
	Cvar_RegisterVariable (&coop, NULL);

	Cvar_RegisterVariable (&pausable, NULL);

	Cvar_RegisterVariable (&temp1, NULL);
	Cvar_RegisterVariable (&temp2, NULL);
	Cvar_RegisterVariable (&temp3, NULL);
	Cvar_RegisterVariable (&temp4, NULL);

	Cvar_RegisterVariable (&test1, NULL);
	Cvar_RegisterVariable (&test2, NULL);
	Cvar_RegisterVariable (&test3, NULL);
	Cvar_RegisterVariable (&test4, NULL);

	Host_FindMaxClients ();
}


/*
===============
Host_WriteConfiguration

Writes key bindings and archived cvars to config.cfg
===============
*/
void Cmd_WriteAlias (FILE *f);

void Host_WriteConfiguration (void)
{
	FILE	*f;

	// dedicated servers initialize the host but don't parse and set the
	// config.cfg cvars
	if (host_initialized & !isDedicated)
	{
		f = fopen (va ("%s/config.cfg", com_gamedir), "w");

		if (!f)
		{
			Con_Printf ("Couldn't write config.cfg.\n");
			return;
		}

		VID_SyncCvars (); // johnfitz -- write actual current mode to config file, in case cvars were messed with

		Key_WriteBindings (f);
		Cvar_WriteVariables (f);
		Cmd_WriteAlias (f);

		// johnfitz -- extra commands to preserve state
		fprintf (f, "vid_restart\n");

		if (in_mlook.state & 1) fprintf (f, "+mlook\n");

		// johnfitz

		fclose (f);

		// johnfitz -- also save fitzquake.rc
#if 0
		f = fopen (va ("%s/fitzquake.rc", GAMENAME), "w"); // always save in id1

		if (!f)
		{
			Con_Printf ("Couldn't write fitzquake.rc.\n");
			return;
		}

		Cvar_WriteVariables (f);
		fprintf (f, "vid_restart\n");

		if (in_mlook.state & 1) fprintf (f, "+mlook\n");

		fclose (f);
#endif
		// johnfitz
	}
}


/*
=================
SV_ClientPrintf

Sends text across to be displayed
FIXME: make this just a stuffed echo?
=================
*/
void SV_ClientPrintf (const char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];

	va_start (argptr, fmt);
	vsprintf (string, fmt, argptr);
	va_end (argptr);

	MSG_WriteByte (&host_client->message, svc_print);
	MSG_WriteString (&host_client->message, string);
}

/*
=================
SV_BroadcastPrintf

Sends text to all active clients
=================
*/
void SV_BroadcastPrintf (const char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];
	int			i;

	va_start (argptr, fmt);
	vsprintf (string, fmt, argptr);
	va_end (argptr);

	for (i = 0; i < svs.maxclients; i++)
		if (svs.clients[i].active && svs.clients[i].spawned)
		{
			MSG_WriteByte (&svs.clients[i].message, svc_print);
			MSG_WriteString (&svs.clients[i].message, string);
		}
}

/*
=================
Host_ClientCommands

Send text over to the client to be executed
=================
*/
void Host_ClientCommands (const char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];

	va_start (argptr, fmt);
	vsprintf (string, fmt, argptr);
	va_end (argptr);

	MSG_WriteByte (&host_client->message, svc_stufftext);
	MSG_WriteString (&host_client->message, string);
}

/*
=====================
SV_DropClient

Called when the player is getting totally kicked off the host
if (crash = true), don't bother sending signofs
=====================
*/
void SV_DropClient (qboolean crash)
{
	int		saveSelf;
	int		i;
	client_t *client;

	if (!crash)
	{
		// send any final messages (don't check for errors)
		if (NET_CanSendMessage (host_client->netconnection))
		{
			MSG_WriteByte (&host_client->message, svc_disconnect);
			NET_SendMessage (host_client->netconnection, &host_client->message);
		}

		if (host_client->edict && host_client->spawned)
		{
			// call the prog function for removing a client
			// this will set the body to a dead frame, among other things
			saveSelf = pr_progs->global_struct->self;
			pr_progs->global_struct->self = EDICT_TO_PROG (host_client->edict);
			PR_ExecuteProgram (&sv_progs, pr_progs->global_struct->ClientDisconnect);
			pr_progs->global_struct->self = saveSelf;
		}

		Sys_Printf ("Client %s removed\n", host_client->name);
	}

	// break the net connection
	NET_Close (host_client->netconnection);
	host_client->netconnection = NULL;

	// free the client (the body stays around)
	host_client->active = false;
	host_client->name[0] = 0;
	host_client->old_frags = -999999;
	net_activeconnections--;

	// send notification to all clients
	for (i = 0, client = svs.clients; i < svs.maxclients; i++, client++)
	{
		if (!client->active)
			continue;

		MSG_WriteByte (&client->message, svc_updatename);
		MSG_WriteByte (&client->message, host_client - svs.clients);
		MSG_WriteString (&client->message, "");
		MSG_WriteByte (&client->message, svc_updatefrags);
		MSG_WriteByte (&client->message, host_client - svs.clients);
		MSG_WriteShort (&client->message, 0);
		MSG_WriteByte (&client->message, svc_updatecolors);
		MSG_WriteByte (&client->message, host_client - svs.clients);
		MSG_WriteByte (&client->message, 0);
	}
}

/*
==================
Host_ShutdownServer

This only happens at the end of a game, not between levels
==================
*/
void Host_ShutdownServer (qboolean crash)
{
	int		i;
	int		count;
	sizebuf_t	buf;
	byte		message[4];
	int		start;

	if (!sv.active)
		return;

	sv.active = false;

	// stop all client sounds immediately
	if (cls.state == ca_connected)
		CL_Disconnect ();

	// flush any pending messages - like the score!!!
	start = SDL_GetTicks ();

	do
	{
		count = 0;

		for (i = 0, host_client = svs.clients; i < svs.maxclients; i++, host_client++)
		{
			if (host_client->active && host_client->message.cursize)
			{
				if (NET_CanSendMessage (host_client->netconnection))
				{
					NET_SendMessage (host_client->netconnection, &host_client->message);
					SZ_Clear (&host_client->message);
				}
				else
				{
					NET_GetMessage (host_client->netconnection);
					count++;
				}
			}
		}

		if ((SDL_GetTicks () - start) > 3000)
			break;
	} while (count);

	// make sure all the clients know we're disconnecting
	buf.data = message;
	buf.maxsize = 4;
	buf.cursize = 0;
	MSG_WriteByte (&buf, svc_disconnect);
	count = NET_SendToAll (&buf, 5);

	if (count)
		Con_Printf ("Host_ShutdownServer: NET_SendToAll failed for %u clients\n", count);

	for (i = 0, host_client = svs.clients; i < svs.maxclients; i++, host_client++)
		if (host_client->active)
			SV_DropClient (crash);

	// clear structures
	memset (&sv, 0, sizeof (sv));
	memset (svs.clients, 0, svs.maxclientslimit * sizeof (client_t));
}


/*
================
Host_ClearMemory

This clears all the memory used by both the client and server, but does
not reinitialize anything.
================
*/
void Host_ClearMemory (void)
{
	Con_DPrintf ("Clearing memory\n");
	D_FlushCaches ();
	Mod_ClearAll ();

	if (host_hunklevel)
		Hunk_FreeToLowMark (host_hunklevel);

	cls.signon = 0;
	memset (&sv, 0, sizeof (sv));
	memset (&cl, 0, sizeof (cl));
}


// ==============================================================================
//
// Host Frame
//
// ==============================================================================

/*
===================
Host_FilterTime

Returns false if the time is too short to run a frame
===================
*/
extern qboolean scr_incapture;
extern cvar_t scr_capturefps;

qboolean Host_FilterTime (unsigned int msec)
{
	static int milliseconds = 0;

	milliseconds += msec;
	realtime = (double) milliseconds * 0.001;

	if (realtime < nextframetime) return false;

	if (cls.timedemo)
		nextframetime = realtime;
	else if (scr_incapture)
		nextframetime += 1.0 / scr_capturefps.value;
	else if (host_maxfps.value > 1000)
		nextframetime = realtime;
	else if (host_maxfps.value < 10)
		nextframetime += 0.1;
	else nextframetime += (1.0 / host_maxfps.value);

	host_frametime = (realtime - oldrealtime);
	oldrealtime = realtime;

	// johnfitz -- host_timescale is more intuitive than host_framerate
	if (host_timescale.value > 0)
		host_frametime *= host_timescale.value;
	// johnfitz
	else if (host_framerate.value > 0)
		host_frametime = host_framerate.value;
	else
	{
		// mh - at over 1000 FPS we don't want to be low-clamping frametime
		if (host_frametime > 0.1) host_frametime = 0.1;
	}

	return true;
}


/*
===================
Host_GetConsoleCommands

Add them exactly as if they had been typed at the console
===================
*/
void Host_GetConsoleCommands (void)
{
	char	*cmd;

	while (1)
	{
		cmd = Sys_ConsoleInput ();

		if (!cmd)
			break;

		Cbuf_AddText (cmd);
	}
}


/*
==================
SV_UpdateServer
==================
*/
void SV_UpdateServer (void)
{
	// use the correct vm
	PR_SetActiveProgs (&sv_progs);

	// run the world state
	pr_progs->global_struct->frametime = sv.frametime;

	// set the time and clear the general datagram
	SV_ClearDatagram ();

	// check for new clients
	SV_CheckForNewClients ();

	// read client messages
	SV_RunClients (sv.frametime);

	// move things around and think
	// always pause in single player if in console or menus
	if (!sv.paused && (svs.maxclients > 1 || key_dest == key_game))
		SV_Physics (sv.frametime);

	// send all messages to the clients
	SV_SendClientMessages ();
}


/*
==================
Host_Frame

Runs all active servers
==================
*/
void CL_AccumulateCmd (double frametime);

extern int scr_framecount;

extern qboolean sys_skipscreenupdate;

void _Host_Frame (unsigned time)
{
	int interframe;
	static int beginframe, endframe;
	int preserver, postserver;
	int pregfx, postgfx;
	int presound, postsound;

	qboolean sv_runframe = false;

	if (host_speeds.value) beginframe = SDL_GetTicks ();

	// something bad happened, or the server disconnected
	if (setjmp (host_abortserver)) return;

	// keep the random time dependent
	if (!cls.demoplayback) rand ();

	// always accumulate commands even if we're not running a frame
	CL_AccumulateCmd ((double) time * 0.001);

	// decide the simulation time
	if (!Host_FilterTime (time)) return;

	interframe = beginframe - endframe;
	if (host_speeds.value) preserver = SDL_GetTicks ();

	// process console commands
	Cbuf_Execute ();
	NET_Poll ();

	sv.accumulatedtime += host_frametime;
	sv_runframe = false;

	// decide if we're going to run a server frame
	if (sv.accumulatedtime >= SV_DELTATIME)
	{
		sv.frametime = sv.accumulatedtime;
		sv.accumulatedtime = 0;
		sv_runframe = true;
	}

	// if running the server locally, make intentions now
	if (sv.active && sv_runframe)
	{
		CL_SendCmd ();
		SV_UpdateServer ();
	}
	else if (!sv.active)
		CL_SendCmd ();

	if (host_speeds.value)
	{
		postserver = SDL_GetTicks ();
		pregfx = postserver;
	}

	// fetch results from server
	if (cls.state == ca_connected)
	{
		if ((sv.active && sv_runframe) || !sv.active)
			CL_UpdateClient (host_frametime, true);
		else CL_UpdateClient (host_frametime, false);
	}

	// update video
	if (!sys_skipscreenupdate)
		SCR_UpdateScreen ();
	else sys_skipscreenupdate = false;

	CL_RunParticles (); // johnfitz -- seperated from rendering

	if (host_speeds.value)
	{
		postgfx = SDL_GetTicks ();
		presound = postgfx;
	}

	// update audio
	if (sv_runframe)
	{
		if (cls.signon == SIGNONS)
			S_Update (r_origin, vpn, vright, vup);
		else S_Update (vec3_origin, vec3_origin, vec3_origin, vec3_origin);

		CDAudio_Update ();
	}

	if (host_speeds.value) postsound = SDL_GetTicks ();

	// fixme - shouldn't this be elsewhere?
	if (cls.signon == SIGNONS)
		CL_DecayLights ();

	if (host_speeds.value) endframe = SDL_GetTicks ();

	// host_speeds is only meaningful if a server frame is run
	if (host_speeds.value && sv_runframe)
	{
		if (key_dest == key_game)
		{
			Con_Printf ("%3i tot  %3i server  %3i  gfx  %3i snd  %3i junk\n",
				endframe - beginframe,
				postserver - preserver,
				postgfx - pregfx,
				postsound - presound,
				interframe);
		}
	}

	host_framecount++;
	host_lastframetime = host_frametime;
}

void Host_Frame (unsigned time)
{
	int time1, time2;
	static int timetotal;
	static int timecount;
	static int timemin = 0x7fffffff;
	static int timemax = 0;
	int i, c, m;

	if (!serverprofile.value)
	{
		_Host_Frame (time);
		return;
	}

	time1 = SDL_GetTicks ();
	_Host_Frame (time);
	time2 = SDL_GetTicks ();

	timetotal += (time2 - time1);
	timecount++;

	if ((time2 - time1) < timemin) timemin = (time2 - time1);
	if ((time2 - time1) > timemax) timemax = (time2 - time1);

	if (timetotal < 1000)
		return;

	m = (timetotal + (timecount >> 1)) / timecount;
	timecount = 0;
	timetotal = 0;
	c = 0;

	for (i = 0; i < svs.maxclients; i++)
	{
		if (svs.clients[i].active)
			c++;
	}

	Con_Printf ("serverprofile: %2i clients  %3i avg msec  %3i min msec  %3i max msec\n", c, m, timemin, timemax);

	timemin = 0x7fffffff;
	timemax = 0;
}

/*
====================
Host_Init
====================
*/
void Sys_Init (void);

void Host_Init (quakeparms_t *parms)
{
	if (standard_quake)
		minimum_memory = MINIMUM_MEMORY;
	else
		minimum_memory = MINIMUM_MEMORY_LEVELPAK;

	if (COM_CheckParm ("-minmemory"))
		parms->memsize = minimum_memory;

	host_parms = *parms;

	if (parms->memsize < minimum_memory)
		Sys_Error ("Only %4.1f megs of memory available, can't execute game", parms->memsize / (float) 0x100000);

	com_argc = parms->argc;
	com_argv = parms->argv;

	Memory_Init (parms->membase, parms->memsize);
	Cbuf_Init ();
	Cmd_Init ();
	LOG_Init (parms);
	Cvar_Init (); // johnfitz
	Sys_Init ();	// for sys cvars that must come up
	V_Init ();
	Chase_Init ();
	COM_Init (parms->basedir);
	Host_InitLocal ();
	W_LoadWadFile (); // johnfitz -- filename is now hard-coded for honesty
	Key_Init ();
	Con_Init ();
	M_Init ();
	PR_Init ();
	Mod_Init ();
	NET_Init ();
	SV_Init ();
	ExtraMaps_Init (); // johnfitz
	Modlist_Init (); // johnfitz

	Con_Printf ("Exe: "__TIME__" "__DATE__"\n");
	Con_Printf ("%4.1f megabyte heap\n", parms->memsize / (1024 * 1024.0));

	if (cls.state != ca_dedicated)
	{
		host_colormap = (byte *) COM_LoadHunkFile ("gfx/colormap.lmp");

		if (!host_colormap)
			Sys_Error ("Couldn't load gfx/colormap.lmp");

		VID_Init ();
		IN_Init (); // moved here, SDL inits the input system with the video system -- kristian
		TM_Init (); // johnfitz
		Draw_Init ();
		SCR_Init ();
		R_Init ();
		S_Init ();
		CDAudio_Init ();
		Sbar_Init ();
		CL_Init ();
	}

	Cbuf_InsertText ("exec quake.rc\n");
	// 	Cbuf_InsertText ("exec fitzquake.rc\n"); // johnfitz (inserted second so it'll be executed first)

	Cbuf_AddText ("\n\nvid_unlock\n"); // johnfitz -- in case the vid mode was locked during vid_init, we can unlock it now.
	// note: added two newlines to the front becuase the command buffer swallows one of them.

	Hunk_AllocName (0, "-HOST_HUNKLEVEL-");
	host_hunklevel = Hunk_LowMark ();

	host_initialized = true;

	Con_Printf ("\n========= Quake Initialized =========\n\n"); // johnfitz - was Sys_Printf
}


/*
===============
Host_Shutdown

FIXME: this is a callback from Sys_Quit and Sys_Error.  It would be better
to run quit through here before the final handoff to the sys code.
===============
*/
void History_Shutdown (void);

void Host_Shutdown (void)
{
	static qboolean isdown = false;

	if (isdown)
	{
		printf ("recursive shutdown\n");
		return;
	}

	isdown = true;

	// keep Con_Printf from trying to update the screen
	scr_disabled_for_loading = true;

	Host_WriteConfiguration ();

	// gb, command history, Baker
	if (con_initialized)
		History_Shutdown ();

	CDAudio_Shutdown ();
	NET_Shutdown ();
	S_Shutdown ();

	if (cls.state != ca_dedicated)
	{
		IN_Shutdown (); // input is only initialized in Host_Init if we're not dedicated -- kristian
		VID_Shutdown ();
	}

	LOG_Close ();
}

